@@ -1,36 +1,50 @@
 """Halogen schema primitives."""
-
-import sys
 import inspect
+from collections import OrderedDict, namedtuple
 
-try:
-    from collections import OrderedDict
-except ImportError:
-    from ordereddict import OrderedDict  # noqa
+from cached_property import cached_property
+import six
 
 from halogen import types
 from halogen import exceptions
 
-PY2 = sys.version_info[0] == 2
 
-if not PY2:
+if not six.PY2:  # pragma: no cover
     string_types = (str,)
-else:
+else:  # pragma: no cover
     string_types = (str, unicode)
 
 
-BYPASS = lambda value: value
+def BYPASS(value):
+    """Bypass getter."""
+    return value
 
 
-def _get_context(func, kwargs):
+ArgSpec = namedtuple("ArgSpec", ["args", "has_kwargs"])
+
+
+if six.PY2:
+
+    def getargspec(function):
+        spec = inspect.getargspec(function)
+        return ArgSpec(args=spec.args, has_kwargs=spec.keywords is not None)
+
+
+else:
+
+    def getargspec(function):
+        spec = inspect.getfullargspec(function)
+        return ArgSpec(args=spec.args + spec.kwonlyargs, has_kwargs=spec.varkw is not None)
+
+
+def _get_context(argspec, kwargs):
     """Prepare a context for the serialization.
 
-    :param func: Function which needs or does not need kwargs.
+    :param argspec: The argspec of the serialization function.
     :param kwargs: Dict with context
     :return: Keywords arguments that function can accept.
     """
-    argspec = inspect.getargspec(func)
-    if argspec.keywords is not None:
+    if argspec.has_kwargs is not None:
         return kwargs
     return dict((arg, kwargs[arg]) for arg in argspec.args if arg in kwargs)
 
@@ -44,6 +58,10 @@ class Accessor(object):
         self.getter = getter
         self.setter = setter
 
+    @cached_property  # Purposefully caching the function signature
+    def _getter_argspec(self):
+        return getargspec(self.getter)
+
     def get(self, obj, **kwargs):
         """Get an attribute from a value.
 
@@ -52,7 +70,7 @@ class Accessor(object):
         """
         assert self.getter is not None, "Getter accessor is not specified."
         if callable(self.getter):
-            return self.getter(obj, **_get_context(self.getter, kwargs))
+            return self.getter(obj, **_get_context(self._getter_argspec, kwargs))
 
         assert isinstance(self.getter, string_types), "Accessor must be a function or a dot-separated string."
 
@@ -94,15 +112,10 @@ class Accessor(object):
 
     def __repr__(self):
         """Accessor representation."""
-        return "<{0} getter='{1}', setter='{1}'>".format(
-            self.__class__.__name__,
-            self.getter,
-            self.setter,
-        )
+        return "<{0} getter='{1}', setter='{2}'>".format(self.__class__.__name__, self.getter, self.setter)
 
 
 class Attr(object):
-
     """Schema attribute."""
 
     creation_counter = 0
@@ -134,7 +147,7 @@ class Attr(object):
         """The key of the this attribute will be placed into (within it's compartment)."""
         return self.name
 
-    @property
+    @cached_property
     def accessor(self):
         """Get an attribute's accessor with the getter and the setter.
 
@@ -149,6 +162,10 @@ class Attr(object):
         attr = self.attr or self.name
         return Accessor(getter=attr, setter=attr)
 
+    @cached_property
+    def _attr_type_serialize_argspec(self):
+        return getargspec(self.attr_type.serialize)
+
     def serialize(self, value, **kwargs):
         """Serialize the attribute of the input data.
 
@@ -168,11 +185,11 @@ class Attr(object):
                     raise
                 value = self.default() if callable(self.default) else self.default
 
-            return self.attr_type.serialize(value, **_get_context(self.attr_type.serialize, kwargs))
+            return self.attr_type.serialize(value, **_get_context(self._attr_type_serialize_argspec, kwargs))
 
         return self.attr_type
 
-    def deserialize(self, value):
+    def deserialize(self, value, **kwargs):
         """Deserialize the attribute from a HAL structure.
 
         Get the value from the HAL structure from the attribute's compartment
@@ -190,28 +207,65 @@ class Attr(object):
             compartment = value[self.compartment]
 
         try:
-            value = self.accessor.get(compartment)
-        except KeyError:
-            if hasattr(self, "default"):
-                value = self.default
-            else:
+            value = self.accessor.get(compartment, **kwargs)
+        except (KeyError, AttributeError):
+            if not hasattr(self, "default") and self.required:
                 raise
+            return self.default() if callable(self.default) else self.default
 
-        return self.attr_type.deserialize(value)
+        return self.attr_type.deserialize(value, **kwargs)
 
     def __repr__(self):
         """Attribute representation."""
-        return "<{0} '{1}'>".format(
-            self.__class__.__name__,
-            self.name,
-        )
+        return "<{0} '{1}'>".format(self.__class__.__name__, self.name)
+
+    def setter(self, setter):
+        """Set an attribute setter accessor function.
+
+        Can be used as a decorator:
+            @total.setter
+            def set_total(obj, value):
+                obj.total = value
+        """
+        self.accessor.setter = setter
+
+    def __call__(self, getter):
+        """Decorate a getter accessor function."""
+        self.name = getter.__name__
+        self.accessor.getter = getter
+        return self
+
+
+def attr(*args, **kwargs):
+    """Attribute as a decorator alias.
+
+    Decorates the getter function like:
+
+        @halogen.attr(AmountType(), default=None)
+        def total(obj):
+            return sum((item.amount for item in obj.items), 0)
+
+    This is identical to using attr with a lambda, but more practical in case of larger functions:
+        total = halogen.Attr(AmountType(), default=None, attr=lambda obj: sum((item.amount for item in obj.items), 0))
+    """
+    return Attr(*args, **kwargs)
 
 
 class Link(Attr):
 
     """Link attribute of a schema."""
 
-    def __init__(self, attr_type=None, attr=None, key=None, required=True, curie=None, templated=None, type=None):
+    def __init__(
+        self,
+        attr_type=None,
+        attr=None,
+        key=None,
+        required=True,
+        curie=None,
+        templated=None,
+        type=None,
+        deprecation=None,
+    ):
         """Link constructor.
 
         :param attr_type: Type, Schema or constant that does the type conversion of the attribute.
@@ -220,6 +274,7 @@ class Link(Attr):
         :param required: Is this link required to be present.
         :param curie: Link namespace prefix (e.g. "<prefix>:<name>") or Curie object.
         :param templated: Is this link templated.
+        :param deprecation: Link deprecation URL.
         :param type: Its value is a string used as a hint to indicate the media type expected when dereferencing
                            the target resource.
         """
@@ -229,19 +284,23 @@ class Link(Attr):
                 attr = BYPASS
 
             attrs = {
-                'templated': templated,
-                'type': type,
+                "templated": templated,
+                "type": type,
+                "deprecation": deprecation,
             }
 
             class LinkSchema(Schema):
                 href = Attr(attr_type=attr_type, attr=BYPASS)
 
-                if attrs['templated'] is not None:
+                if attrs["templated"] is not None:
                     templated = Attr(attr=lambda value: templated)
 
-                if attrs['type'] is not None:
+                if attrs["type"] is not None:
                     type = Attr(attr=lambda value: type)
 
+                if attrs["deprecation"] is not None:
+                    deprecation = Attr(attr=lambda value: deprecation)
+
             attr_type = LinkSchema
 
         super(Link, self).__init__(attr_type=attr_type, attr=attr, required=required)
@@ -340,20 +399,20 @@ class _Schema(types.Type):
 
     def __new__(cls, **kwargs):
         """Create schema from keyword arguments."""
-        schema = type("Schema", (cls, ), {"__doc__": cls.__doc__})
-        schema.__class_attrs__ = []
-        schema.__attrs__ = []
+        schema = type("Schema", (cls,), {"__doc__": cls.__doc__})
+        schema.__class_attrs__ = OrderedDict()
+        schema.__attrs__ = OrderedDict()
         for name, attr in kwargs.items():
             if not hasattr(attr, "name"):
                 attr.name = name
-            schema.__class_attrs__.append(attr)
-            schema.__attrs__.append(attr)
+            schema.__class_attrs__[attr.name] = attr
+            schema.__attrs__[attr.name] = attr
         return schema
 
     @classmethod
     def serialize(cls, value, **kwargs):
         result = OrderedDict()
-        for attr in cls.__attrs__:
+        for attr in cls.__attrs__.values():
             compartment = result
             if attr.compartment is not None:
                 compartment = result.setdefault(attr.compartment, OrderedDict())
@@ -367,7 +426,7 @@ class _Schema(types.Type):
         return result
 
     @classmethod
-    def deserialize(cls, value, output=None):
+    def deserialize(cls, value, output=None, **kwargs):
         """Deserialize the HAL structure into the output value.
 
         :param value: Dict of already loaded json which will be deserialized by schema attributes.
@@ -375,30 +434,31 @@ class _Schema(types.Type):
 
         :returns: Dict of deserialized value for attributes. Where key is name of schema's attribute and value is
         deserialized value from value dict.
+        :raises: ValidationError.
         """
         errors = []
         result = {}
-        for attr in cls.__attrs__:
+        for attr in cls.__attrs__.values():
             try:
-                result[attr.name] = attr.deserialize(value)
+                result[attr.name] = attr.deserialize(value, **kwargs)
             except NotImplementedError:
                 # Links don't support deserialization
                 continue
+            except ValueError as e:
+                errors.append(exceptions.ValidationError(e, attr.name))
             except exceptions.ValidationError as e:
                 e.attr = attr.name
                 errors.append(e)
-            except KeyError:
+            except (KeyError, AttributeError):
                 if attr.required:
-                    e = exceptions.ValidationError("Missing attribute.", attr.name)
-                    e.attr = attr.name
-                    errors.append(e)
+                    errors.append(exceptions.ValidationError("Missing attribute.", attr.name))
 
         if errors:
             raise exceptions.ValidationError(errors)
 
         if output is None:
             return result
-        for attr in cls.__attrs__:
+        for attr in cls.__attrs__.values():
             if attr.name in result:
                 attr.accessor.set(output, result[attr.name])
 
@@ -409,7 +469,7 @@ class _SchemaType(type):
 
     def __init__(cls, name, bases, clsattrs):
         """Create a new _SchemaType."""
-        cls.__class_attrs__ = []
+        cls.__class_attrs__ = OrderedDict()
         curies = set([])
 
         attrs = [(key, value) for key, value in clsattrs.items() if isinstance(value, Attr)]
@@ -418,7 +478,7 @@ class _SchemaType(type):
         # Collect the attributes and set their names.
         for name, attr in attrs:
             delattr(cls, name)
-            cls.__class_attrs__.append(attr)
+            cls.__class_attrs__[name] = attr
             if not hasattr(attr, "name"):
                 attr.name = name
 
@@ -431,23 +491,18 @@ class _SchemaType(type):
 
         if curies:
             link = LinkList(
-                Schema(
-                    href=Attr(),
-                    name=Attr(),
-                    templated=Attr(required=False),
-                    type=Attr(required=False),
-                ),
+                Schema(href=Attr(), name=Attr(), templated=Attr(required=False), type=Attr(required=False)),
                 attr=lambda value: list(curies),
                 required=False,
             )
             link.name = "curies"
 
-            cls.__class_attrs__.append(link)
+            cls.__class_attrs__[link.name] = link
 
-        cls.__attrs__ = []
+        cls.__attrs__ = OrderedDict()
         for base in reversed(cls.__mro__):
-            cls.__attrs__.extend(getattr(base, "__class_attrs__", []))
+            cls.__attrs__.update(getattr(base, "__class_attrs__", OrderedDict()))
 
 
-Schema = _SchemaType("Schema", (_Schema, ), {"__doc__": _Schema.__doc__})
+Schema = _SchemaType("Schema", (_Schema,), {"__doc__": _Schema.__doc__})
 """Schema is the basic class used for setting up schemas."""
